///////////////////////////////////////////////////
//Based on code provided by Steven Theriault, and
//the Atmel TWI sample code.  
//
//sonar.c
///////////////////////////////////////////////////
#include <io.h>
#include "global.h"
#include <twi.h>

#define TWI_RATE_SELECT 21   //100kHz clock @ 6MHz fck

#define SONAR_ADDRESS 0xE0
#define READ 1
#define WRITE 0

#define SONAR_COMMAND_REGISTER 0
#define SONAR_COMMAND 80 //result in inches
#define SONAR_MAX_ADDRESS 35 //last byte available (36 bytes 0:35)



void sonar_init(void)
{
    PORTC &= ~(_BV(1)|_BV(0));  //SDA, SCL are outputs, 
    DDRC |= _BV(1)|_BV(0);

    TWBR = TWI_RATE_SELECT;          //set clock 
    TWCR = _BV(TWEA) | _BV(TWEN) | _BV(TWINT); //acknoledge, enable, clear IRQ
    TWAR = 0x01;        // Don't care, won't be in Slave mode
}

//start the sonar "ping"
//note that it takes at least 65ms for the data to 
//be available (so it's best to do something else
//while the sonar is processing)
s08 sonar_start(void)
{
    loop_until_bit_is_clear(TWCR, TWSTO);  //check no transmission in progress
    
    //based on demo code on p. 112 in mega323 manual
    TWCR = _BV(TWSTA) | _BV(TWEN); //send start condition
    while(!(TWCR & _BV(TWINT))){;} //wait for interrupt flag to get set
    if(TWSR != TW_START) {
        TWCR = _BV(TWINT)|_BV(TWSTO)|_BV(TWEN); 
        return -1;  //error occured - should have been start
    } 
    
    TWDR = SONAR_ADDRESS | WRITE;  //send sonar address
    TWCR = _BV(TWINT) | _BV(TWEN); //clear interrupt to start transmission
    while(!(TWCR & _BV(TWINT))){;} //wait for interrupt flag to get set
    if(TWSR != TW_MT_SLA_ACK) {
        TWCR = _BV(TWINT)|_BV(TWSTO)|_BV(TWEN); 
        return -2;
    }
    
    TWDR = SONAR_COMMAND_REGISTER;  //send the address on the sonar to write to
    TWCR = _BV(TWINT) | _BV(TWEN); //clear IRQ, continue transmission
    while(!(TWCR & _BV(TWINT))){;} //wait for interrupt flag to get set
    if(TWSR != TW_MT_DATA_ACK) {
        TWCR = _BV(TWINT)|_BV(TWSTO)|_BV(TWEN); 
        return -3;
    }
    
    TWDR = SONAR_COMMAND;          //send the sonar "start ranging" command
    TWCR = _BV(TWINT) | _BV(TWEN);
    while(!(TWCR & _BV(TWINT))){;} //wait for interrupt flag to get set
    if(TWSR != TW_MT_DATA_ACK){
        TWCR = _BV(TWINT)|_BV(TWSTO)|_BV(TWEN); 
        return -4;
    }
    
    TWCR = _BV(TWINT) | _BV(TWSTO) | _BV(TWEN); //send stop condition
    
    return 0;
    
}

//data should be 36 long
s08 sonar_get_data(u08* data) 
{
    //char out[32];
    //u08* ranges = (u08*) _ranges;
    loop_until_bit_is_clear(TWCR, TWSTO); //wait for last transmission to end
    
    TWCR = _BV(TWINT)|_BV(TWSTA)|_BV(TWEN); //send start condition
    loop_until_bit_is_set(TWCR, TWINT);
    if(TWSR != TW_START) {
        TWCR = _BV(TWINT)|_BV(TWSTO)|_BV(TWEN); 
        //sprintf(out, "1: %x", TWSR);      
        //uart_putstr(out); 
        return -1;
        //return TWSR;
    }
    
    TWDR = SONAR_ADDRESS | WRITE;         //send sonar address
    TWCR = _BV(TWINT)|_BV(TWEN);
    loop_until_bit_is_set(TWCR, TWINT);
    if(TWSR != TW_MT_SLA_ACK) {
        TWCR = _BV(TWINT)|_BV(TWSTO)|_BV(TWEN); 
        //sprintf(out, "2: %x", TWSR);
        //uart_putstr(out);
        return 1;            //expected error when sonar is busy
    }
    
    TWDR = 0;    //send address to start reading at (register 0)
    TWCR = _BV(TWINT)|_BV(TWEN);
    loop_until_bit_is_set(TWCR, TWINT);
    if(TWSR != TW_MT_DATA_ACK) {
        TWCR = _BV(TWINT)|_BV(TWSTO)|_BV(TWEN); 
        //sprintf(out, "3: %x", TWSR);
        //uart_putstr(out);
        return 1;                  //expected error when sonar is busy
    }
    
    TWCR = _BV(TWINT)|_BV(TWSTA)|_BV(TWEN);  //send repeated start
    loop_until_bit_is_set(TWCR, TWINT);
    if(TWSR != TW_REP_START) {
        TWCR = _BV(TWINT)|_BV(TWSTO)|_BV(TWEN);
        //sprintf(out, "4: %x", TWSR);
        //uart_putstr(out);
        return -4;
    }
    
    TWDR = SONAR_ADDRESS | READ;            //send sonar address for read
    TWCR = _BV(TWINT)|_BV(TWEA)|_BV(TWEN);
    loop_until_bit_is_set(TWCR, TWINT);
    if(TWSR != TW_MR_SLA_ACK) {
        TWCR = _BV(TWINT)|_BV(TWSTO)|_BV(TWEN); 
        return -5;
    }
    
    u08 i;
    for(i=0; i<SONAR_MAX_ADDRESS; i++) {
        TWCR = _BV(TWINT)|_BV(TWEA)|_BV(TWEN);
        loop_until_bit_is_set(TWCR, TWINT);
        if(TWSR != TW_MR_DATA_ACK) {
            TWCR = _BV(TWINT)|_BV(TWSTO)|_BV(TWEN); 
            return -6;
        }
        *data = TWDR;
        data++;
    }
    
    TWCR = _BV(TWINT)|_BV(TWEN);   //TWEA not set - to indicate this is last byte
    loop_until_bit_is_set(TWCR, TWINT);
    if(TWSR != TW_MR_DATA_NACK) {   //get NACK since we didn't set TWEA
        TWCR = _BV(TWINT)|_BV(TWSTO)|_BV(TWEN); 
        return -7;
    }
    *data = TWDR;  //get last byte
    
    TWCR = _BV(TWINT)|_BV(TWSTO)|_BV(TWEN);  //send stop condition
    
    return 0;
}


//data is input from sonar_get_data, ranges and light_sensor are output
//data[36], ranges[17], light_sensor[1];
void convert_data(u16* ranges, u08* light_sensor, u08* data) 
{
    data++;
    *light_sensor = *data; //1st element is light sensor
    data++;         //first range data byte (high byte)
    u08 i;
    //step through the 17 range locations
    for(i=0; i<17; i++) {
        *ranges = (*data)<<8 | *(data+1);
        data+=2;
        ranges++;
    }
}

//returns the light sensor value,
//gets the ranges in the provided array
//which must be 17 elements long
s08 sonar_data(u16* ranges, u08* light_sensor) 
{
    u08 data[36];
    s08 error;
    do {
        error = sonar_get_data(data);
    } while(error==1);
    if(error == 0) {
        convert_data(ranges, light_sensor, data);
        return 0;
    } else {
        return -1;
    }    
}

u16 sonar_get_closest(u16* ranges)
{ 
    //things closer than 6inches are probably reflections
    //off the front of the robot
    while(*ranges <= 6 && *ranges!=0) {
        ranges++;
    }
    return *ranges;
}

//returns the closest item to the sonar
//this is the easiest way to use the sonar - 
//just call sonar_start(); then call sonar_closest();
//to get the closest reading.
//returns 0 if an error occurred.  
u16 sonar_closest(void)
{
    u16 ranges[17];
    u08 temp;
    s08 error;
    error = sonar_data(ranges, &temp);
    if(error == 0) {
        return sonar_get_closest(ranges);
    } else {
        return 0;
    }
    
}


/*
int main(void)
{
    DDRA = 0;    //all ports are inputs
    DDRB = 0;
    DDRC = 0; 
    DDRD = 0;
    
    //u08 data[36]; //sonar data
    u16 ranges[17];
    u08 light_sensor=0;
    char out[32];
    
    delay(0x2000); //give bootloader a second to reset
    
    uart_init();
    sonar_init();
    sei(); 
    
    s08 error1, error2;
    u08 i;
    
    for(;;) {
        error1 = sonar_start();
        error2 = sonar_data(ranges, &light_sensor);
        
        if(error1==0 && error2==0) {
            for(i=0; i< 17; i++) {
                if(i%8 == 0) {
                    uart_nl();
                }
                sprintf(out, " %i:%i ",i,ranges[i]);
                uart_putstr(out);
            }
            uart_nl();
        }
        sprintf(out, "light: %i\r\nerror1: %i error2: %i", light_sensor, error1, error2);
        uart_putstr(out);
        uart_nl();
        delay(0x2000);
    }
    
}
*/

