//UART library functions.
//based on sample code by Volker Oth in
//the gcctest9 package of avr-gcc 3.2
//from www.avrfreaks.net
/*
    Title:    UART library
    Author:   Volker Oth
    Date:     5/1999
    Purpose:  Sending a block of data to uart.
    needed
    Software: AVR-GCC to compile, AVA to assemble and link
    needed
    Hardware: ATS90S8515 on STK200 board
    Note:     To contact me, mail to
                  volkeroth@gmx.de
              You might find more AVR related stuff at my homepage:
                  http://members.xoom.com/volkeroth
*/

#include <stdarg.h>
#include <string-avr.h>
#include <io.h>
#include <sig-avr.h>
#include <interrupt.h>
#include "global.h"
//#include "timer.h"
#include "uart.h"


/* uart globals */
volatile u08 uart_txd_buf_cnt;
volatile u08 uart_rxd_buf_cnt;
u08 *uart_txd_in_ptr, *uart_txd_out_ptr;
u08 *uart_rxd_in_ptr, *uart_rxd_out_ptr;
u08 UART_CLR[] = {ESC, '[','H', ESC, '[', '2', 'J',0};
u08 UART_NL[] = {0x0d,0x0a,0};
u08 uart_txd_buffer[UART_BUF_SIZE];
u08 uart_rxd_buffer[UART_BUF_SIZE];

void uart_init(void)
// initialize uart 
{
    UBRRH = 0x00;
    UBRRL = UART_BAUD_SELECT;
    UCSRB = _BV(RXCIE)|_BV(RXEN)|_BV(TXEN);
    UCSRC = _BV(URSEL)|_BV(UCSZ1)|_BV(UCSZ0);
    UCSRA = 0x00;       
    
    uart_txd_in_ptr  = uart_txd_out_ptr = uart_txd_buffer;
    uart_rxd_in_ptr  = uart_rxd_out_ptr = uart_rxd_buffer;
    uart_txd_buf_cnt = 0;
    uart_rxd_buf_cnt = 0;
}


SIGNAL(SIG_UART_DATA)      
/* signal handler for uart data buffer empty interrupt */
{
    if (uart_txd_buf_cnt > 0) {
        outp(*uart_txd_out_ptr, UDR);           /* write byte to data buffer */
        if (++uart_txd_out_ptr >= uart_txd_buffer + UART_BUF_SIZE) /* Pointer wrapping */
            uart_txd_out_ptr = uart_txd_buffer;
        if(--uart_txd_buf_cnt == 0)             /* if buffer is empty: */
            cbi(UCSRB, UDRIE);                    /* disable UDRIE int */
    }
}


SIGNAL(SIG_UART_RECV)      
/* signal handler for receive complete interrupt */
{
    *uart_rxd_in_ptr = inp(UDR);            /* read byte from receive register */
    uart_rxd_buf_cnt++;
    if (++uart_rxd_in_ptr >= uart_rxd_buffer + UART_BUF_SIZE) /* Pointer wrapping */
        uart_rxd_in_ptr = uart_rxd_buffer;
}


s16 uart_getchar(void)
{
    u08 c;

    if (uart_rxd_buf_cnt>0) {
        cli();
        uart_rxd_buf_cnt--;
        c = *uart_rxd_out_ptr;              /* get character from buffer */
        if (++uart_rxd_out_ptr >= uart_rxd_buffer + UART_BUF_SIZE) /* pointer wrapping */
            uart_rxd_out_ptr = uart_rxd_buffer;
        sei();
        return c;
    } 
    else 
        return -1;                          /* buffer is empty */
}


bool uart_putchar(u08 c)
{
    if (uart_txd_buf_cnt<UART_BUF_SIZE) {
        cli();
        uart_txd_buf_cnt++;
        *uart_txd_in_ptr = c;               /* put character into buffer */
        if (++uart_txd_in_ptr >= uart_txd_buffer + UART_BUF_SIZE) /* pointer wrapping */
            uart_txd_in_ptr = uart_txd_buffer;
        sbi(UCSRB, UDRIE);                    /* enable UDRIE int */
        sei();
        return 1;
    } 
    else 
        return 0;                           /* buffer is full */
}


bool uart_putstr(u08 s[])
{
    char *c = s;

    while (*c)
        if (uart_putchar(*c))
            c++;

    return 1;
}

bool uart_putstr_fast(u08 s[])
/*puts as much of the string as possible into the buffer*/
{
    char *c = s;
    bool success=1;
    
    while(*c && success)
    {
        success=uart_putchar(*c);
        c++;
    }
     
    return success;
}


void uart_clr(void)
/* Send a 'clear screen' to a VT100 terminal */
{
    uart_putstr(UART_CLR); 
}


void uart_nl(void)
/* Send a 'new line' */
{
    uart_putstr(UART_NL); 
}



int sprintf(u08 *buf, const u08 *format, ...)
/* simplified sprintf */
{
  u08 scratch[SCRATCH];
  u08 format_flag;
  u16 u_val=0, base;
  u08 *ptr;
  va_list ap;

  va_start (ap, format);
  for (;;){
    while ((format_flag = *format++) != '%'){      /* Until '%' or '\0' */
      if (!format_flag){va_end (ap); return (0);}
      *buf = format_flag; buf++; *buf=0;
    }

    switch (format_flag = *format++){

    case 'c':
      format_flag = va_arg(ap,int);
    default:
      *buf = format_flag; buf++; *buf=0;
      continue;
    case 'S':
    case 's':
      ptr = va_arg(ap,char *);
      strcat(buf, ptr);
      continue;
    case 'o':
      base = 8;
      *buf = '0'; buf++; *buf=0;
      goto CONVERSION_LOOP;
    case 'i':  
      if (((int)u_val) < 0){
        u_val = - u_val;
        *buf = '-'; buf++; *buf=0;
      }
      /* no break -> run into next case */
    case 'u':
      base = 10;
      goto CONVERSION_LOOP;
    case 'x':
      base = 16;

    CONVERSION_LOOP:
      u_val = va_arg(ap,int);
      ptr = scratch + SCRATCH;
      *--ptr = 0;
      do {
        char ch = u_val % base + '0';
        if (ch > '9')
          ch += 'a' - '9' - 1;
        *--ptr = ch;
        u_val /= base;
      } while (u_val);
      strcat(buf, ptr);
      buf += strlen(ptr);
    }
  }
}

